package fullenv

import (
	"bufio"
	"bytes"
	"errors"
	"fmt"
	"io/ioutil"
	"os"
	"os/exec"
	"path"
	"reflect"
	"strings"
	"syscall"
	"testing"
	"time"

	"github.com/spf13/pflag"
	"github.com/spf13/viper"
	"github.com/tgulacsi/go-locking"
	"gopkg.in/yaml.v2"
	// Embed QB dialects we need
	_ "github.com/slicebit/qb/dialects/postgres"

	"bitbucket.org/orus-io/go-xbus"
	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/xbus/storage/sql/model"
	"bitbucket.org/orus-io/xbus/testutils"

	_ "github.com/lib/pq" // full env tests run on postgres
)

// RawSettings holds the raw configuration of a xbus program (xbus, xbus-client...)
type RawSettings yaml.MapSlice

func mapSliceSet(s *RawSettings, key string, value interface{}) {
	for i, item := range *s {
		sKey, ok := item.Key.(string)
		if !ok {
			panic("key is not a string")
		}
		if sKey == key {
			item.Value = value
			(*s)[i] = item
			return
		}
	}
	if !strings.Contains(key, ".") {
		*s = append(*s, yaml.MapItem{Key: key, Value: value})
		return
	}
	splittedKey := strings.SplitN(key, ".", 2)
	root := splittedKey[0]
	path := splittedKey[1]
	for i, item := range *s {
		sKey, _ := item.Key.(string)
		if sKey == root {
			submap, ok := item.Value.(RawSettings)
			if !ok {
				panic(fmt.Sprintf("%s is not a rawsettings: %v", root, reflect.TypeOf(item.Value)))
			}
			mapSliceSet(&submap, path, value)
			item.Value = submap
			(*s)[i] = item
			return
		}
	}
	v := RawSettings{}
	mapSliceSet(&v, path, value)
	*s = append(*s, yaml.MapItem{Key: root, Value: v})
}

// Options holds all the options necessary to setup a FullEnv
type Options struct {
	Xbusd   RawSettings
	Clients []RawSettings
	Graphs  map[string]RawSettings

	Logger xbus.Logger
}

// OptionsFromFile initialize a Options from a yaml configuration file
func OptionsFromFile(configfile string) (Options, error) {
	buf, err := ioutil.ReadFile(configfile)
	if err != nil {
		return Options{}, err
	}

	var options Options
	if err := yaml.Unmarshal(buf, &options); err != nil {
		return options, err
	}

	return options, nil
}

// NewFullEnvForTest returns a FullEnv that sends logs to a *testing.T
func NewFullEnvForTest(t *testing.T, basename, configfile string) *FullEnv {
	options, err := OptionsFromFile(configfile)
	if err != nil {
		t.Fatal(err)
	}
	options.Logger = testutils.GetTestLogger(t)

	env := NewFullEnv(basename, options)
	env.t = t
	return env
}

// NewFullEnv initializes a FullEnv
func NewFullEnv(basename string, options Options) *FullEnv {
	return &FullEnv{
		basename: basename,
		options:  options,
		log:      options.Logger,

		extClients:       make(map[string]*exec.Cmd),
		inClients:        make(map[string]*xbus.Client),
		runningInClients: make(map[string]bool),
	}
}

// FullEnv pilots xbusd and clients
type FullEnv struct {
	basename string
	options  Options

	t       *testing.T
	log     xbus.Logger
	dirname string
	dsn     string
	dblock  *locking.PortLock

	actorServices []string

	xbusd   *exec.Cmd
	XbusCtl CmdRunner

	extClients       map[string]*exec.Cmd
	inClients        map[string]*xbus.Client
	runningInClients map[string]bool

	Emitter  *Emitter
	Consumer *Consumer
}

// RegisterActorService registers an actor service for this env only
// Allow to inject actors that are controlled by the test
func (fe *FullEnv) RegisterActorService(name string, init xbus.NewActorServiceFunc) {
	fe.actorServices = append(fe.actorServices, name)
	xbus.RegisterActorService(name, init)
}

func (fe *FullEnv) setupDB() error {
	dsn, ok := os.LookupEnv("XBUS_TEST_POSTGRES_DSN")
	if !ok {
		return errors.New("Need a XBUS_TEST_POSTGRES_DSN environment variable")
	}

	fe.dblock = locking.NewPortLock(1764)
	err := fe.dblock.Lock()
	if err != nil {
		fe.log.Errorf("Cannot get lock: %s", err)
		return err
	}

	_, db, err := model.NewDB("postgres", dsn, 2, fe.log, 0)
	if err != nil {
		return err
	}

	// Cleanup the db
	tables := db.Metadata.GetQbMetadata().Tables()
	for i := len(tables) - 1; i >= 0; i-- {
		table := tables[i]
		drop := table.Drop(db.Engine.Dialect())
		if _, err := db.Engine.DB().Exec(drop); err != nil {
			fe.log.Noticef("%s", err)
		}
	}

	db.Close()

	fe.dsn = dsn

	mapSliceSet(&fe.options.Xbusd, "database.dsn", dsn)

	return nil
}

func (fe *FullEnv) runXbusd(args ...string) error {
	p := os.Getenv("GOPATH")

	xbusdCfgFilename := path.Join(fe.dirname, "xbusd.yaml")

	args = append([]string{
		"--config", xbusdCfgFilename,
	}, args...)

	cmd := exec.Command(
		path.Join(p, "bin", "xbusd"), args...)
	cmd.Dir = fe.dirname

	b, err := cmd.CombinedOutput()
	if err != nil {
		fe.log.Errorf("%s", b)
		return fmt.Errorf("Could not run xbusd : %s", err)
	}

	return nil
}

func (fe *FullEnv) xbusClientArgs(name string, args ...string) (string, []string) {
	p := os.Getenv("GOPATH")

	xbusclientCfgFilename := path.Join(fe.dirname, "client-"+name, "xbus-client.yaml")

	args = append([]string{
		"--config", xbusclientCfgFilename,
	}, args...)

	return path.Join(p, "bin", "xbus-client"), args
}

func (fe *FullEnv) runXbusClient(name string, args ...string) error {
	program, finalArgs := fe.xbusClientArgs(name, args...)

	cmd := exec.Command(program, finalArgs...)
	cmd.Dir = fe.dirname

	b, err := cmd.CombinedOutput()
	if err != nil {
		fe.log.Errorf("xbus-client: %s", b)
		return fmt.Errorf("Could not run xbus-client : %s", err)
	}

	return nil
}

func (fe *FullEnv) setupServer() error {
	// create the configuration file
	conf, err := yaml.Marshal(fe.options.Xbusd)
	if err != nil {
		panic(err)
	}
	xbusdCfgFilename := path.Join(fe.dirname, "xbusd.yaml")
	if err := ioutil.WriteFile(xbusdCfgFilename, conf, 0600); err != nil {
		return err
	}
	if err := fe.runXbusd("init"); err != nil {
		return err
	}

	if err := fe.runXbusd("key", "generate"); err != nil {
		return err
	}

	if err := fe.runXbusd("cert", "generate", "server-ca"); err != nil {
		return err
	}

	if err := fe.runXbusd("cert", "generate", "client-ca"); err != nil {
		return err
	}

	if err := fe.StartXbusd(); err != nil {
		return err
	}

	return nil
}

func (fe *FullEnv) setupCtl() (err error) {
	defer catcher(&err)

	dirname := path.Join(fe.dirname, "xbusctl")
	must(os.Mkdir(dirname, 0700))
	p := os.Getenv("GOPATH")
	xbusdCtlFilename := path.Join(dirname, "xbusctl.yaml")
	binpath := path.Join(p, "bin", "xbusctl")

	if err := ioutil.WriteFile(xbusdCtlFilename, []byte("account-name: ctl"), 0600); err != nil {
		return err
	}

	xbusctl := NewCmdRunner(
		xbus.NewPrefixedLogger("-- xbusctl --", fe.log),
		dirname, binpath, "--config", xbusdCtlFilename)

	xbusctl.MustRun("key", "generate")
	xbusctl.MustRun("register")

	must(fe.runXbusd("account", "accept", "ctl"))

	xbusctl.MustRun("register")

	fe.XbusCtl = xbusctl
	return nil
}

func (fe *FullEnv) setupGraphs() (err error) {
	defer catcher(&err)

	graphDir := path.Join(fe.dirname, "graphs")
	must(os.Mkdir(graphDir, 0700))

	for name, graphDef := range fe.options.Graphs {
		graph, err := yaml.Marshal(graphDef)
		must(err)

		graphfile := path.Join(graphDir, name+".yaml")
		must(ioutil.WriteFile(graphfile, graph, 0600))
		fe.XbusCtl.MustRun("graph", "save", name, "1", graphfile)
		fe.XbusCtl.MustRun("graph", "status", name, "1", "--active")
	}
	return nil
}

func (fe *FullEnv) setupClient(config RawSettings) error {
	var (
		name      string
		inProcess bool
	)

	for _, item := range config {
		if s, ok := item.Key.(string); ok {
			if s == "inprocess" {
				if b, ok := item.Value.(bool); ok {
					inProcess = b
				}
			}
			if s == "account-name" {
				if s, ok := item.Value.(string); ok {
					name = s
				}
			}
		}
	}

	var (
		// TODO forbid names with path separators
		configDir  = path.Join(fe.dirname, "client-"+name)
		configFile = "xbus-client.yaml"
	)
	if err := os.MkdirAll(configDir, 0700); err != nil {
		return err
	}
	b, err := yaml.Marshal(config)
	if err != nil {
		return err
	}
	if err := ioutil.WriteFile(path.Join(configDir, configFile), b, 0600); err != nil {
		return err
	}

	if inProcess {
		v := viper.New()
		xbus.DefineAndBindPFlags(v, &pflag.FlagSet{})
		v.SetConfigFile(configDir + "/xbus-client.yaml")
		err := v.ReadInConfig()
		if err != nil {
			return err
		}
		coptions := xbus.COptionsFromViper(
			v, api.AccountTypeActor, fe.options.Logger, configDir, configFile, true)
		client := xbus.NewClient(coptions, fe.options.Logger)
		if err := client.CreateKey(false); err != nil {
			return err
		}
		if err := client.CreateCSR("localhost", "", name, false); err != nil {
			return err
		}
		if err := client.Connect(); err != nil {
			return err
		}
		status, err := client.Register()
		if err != nil {
			return err
		}
		if status != api.RegistrationPending {
			return fmt.Errorf("Invalid return from Register: %s (should be 'pending')", status)
		}
		if err := fe.XbusCtl.Run("account", "accept", client.ID().String()); err != nil {
			return err
		}

		for _, actor := range coptions.ActorList {
			if err := fe.XbusCtl.Run("actor", "accept", actor.Name); err != nil {
				return err
			}
		}

		status, err = client.Register()
		if err != nil {
			return err
		}
		if status != api.RegistrationAccepted {
			return fmt.Errorf("Invalid return from Register: %s (should be 'accepted')", status)
		}
		// Reconnect so the new permissions are applied
		client.Close()
		if err := client.Connect(); err != nil {
			return err
		}
		fe.inClients[name] = client
	} else {
		if err := fe.runXbusClient(name, "register"); err != nil {
			return err
		}

		// Load the conf to read the ids
		v := viper.New()
		xbus.DefineAndBindPFlags(v, &pflag.FlagSet{})
		v.SetConfigFile(configDir + "/xbus-client.yaml")
		err := v.ReadInConfig()
		if err != nil {
			return err
		}
		coptions := xbus.COptionsFromViper(
			v, api.AccountTypeActor, fe.options.Logger, configDir, configFile, true)

		// Accept the account and actor registration
		fe.log.Noticef("Accepting account %s", coptions.ID)
		if err := fe.XbusCtl.Run("account", "accept", coptions.ID.String()); err != nil {
			return err
		}

		for _, actor := range coptions.ActorList {
			fe.log.Noticef("Accepting %s", actor.ID)
			if err := fe.XbusCtl.Run("actor", "accept", actor.ID.String()); err != nil {
				return err
			}
		}

		if err := fe.runXbusClient(name, "register"); err != nil {
			return err
		}

		fe.extClients[name] = nil
	}
	return nil
}

func (fe *FullEnv) setupClients() error {
	for _, config := range fe.options.Clients {
		if err := fe.setupClient(config); err != nil {
			return err
		}
	}
	return nil
}

// Init setup the server, the clients and load the graphs
// Once done only the server is running
func (fe *FullEnv) Init() error {
	dirname, err := ioutil.TempDir("", "xbus-fullenv-tests-"+fe.basename+"--")
	if err != nil {
		return err
	}
	if err := os.MkdirAll(path.Join(dirname, "certs", "xbusd"), 0700); err != nil {
		return err
	}

	fe.log.Noticef("Fullenv working dir created: %s", dirname)

	var success bool
	defer func(s *bool) {
		if !*s {
			fe.Close()
		}
	}(&success)
	fe.Emitter = NewEmitter(fe.log)
	fe.Consumer = NewConsumer(fe.log)
	fe.RegisterActorService("fullenv.emitter", fe.Emitter.Service)
	fe.RegisterActorService("fullenv.consumer", fe.Consumer.Service)
	fe.dirname = dirname
	if err := fe.setupDB(); err != nil {
		return err
	}
	if err := fe.setupServer(); err != nil {
		return err
	}
	if err := fe.setupCtl(); err != nil {
		return err
	}
	if err := fe.setupClients(); err != nil {
		return err
	}
	if err := fe.setupGraphs(); err != nil {
		return err
	}
	success = true
	return nil
}

func (fe *FullEnv) startInProcessClient(name string) error {
	client := fe.inClients[name]
	err := client.Startup()
	if err != nil {
		return err
	}
	fe.runningInClients[name] = true
	return nil
}

func (fe *FullEnv) stopInProcessClient(name string) error {
	client := fe.inClients[name]
	client.Shutdown()
	fe.runningInClients[name] = false
	return nil
}

func (fe *FullEnv) startExtProcessClient(name string) error {
	cmd := fe.extClients[name]
	if cmd != nil {
		return fmt.Errorf("Client '%s' aready started", name)
	}
	program, args := fe.xbusClientArgs(name, "serve")
	cmd = exec.Command(program, args...)
	cmd.Dir = fe.dirname

	if err := captureCmdOutputs("client-"+name, cmd, fe.log); err != nil {
		return err
	}

	err := cmd.Start()
	if err != nil {
		return err
	}
	fe.extClients[name] = cmd
	return nil
}

func (fe *FullEnv) stopExtProcessClient(name string) error {
	cmd := fe.extClients[name]
	if cmd == nil {
		return fmt.Errorf("Client '%s' is not running", name)
	}
	cmd.Process.Signal(os.Interrupt)
	var timer *time.Timer
	timer = time.AfterFunc(time.Second, func() {
		fe.log.Errorf("Client '%s' still running. Attempting a SIGQUIT", name)
		if err := cmd.Process.Signal(syscall.SIGQUIT); err != nil {
			fe.log.Errorf("SIGQUIT failed with '%s', killing the process.", err)
			cmd.Process.Kill()
		} else {
			timer = time.AfterFunc(time.Second, func() {
				fe.log.Errorf("Client '%s' still running. Killing it", name)
				cmd.Process.Kill()
			})
		}
	})
	defer timer.Stop()
	err := cmd.Wait()
	fe.extClients[name] = nil
	return err
}

// StartClient starts a client
func (fe *FullEnv) StartClient(name string) error {
	if _, ok := fe.inClients[name]; ok {
		return fe.startInProcessClient(name)
	} else if _, ok := fe.extClients[name]; ok {
		return fe.startExtProcessClient(name)
	} else {
		return fmt.Errorf("No such client: %s", name)
	}
}

// StopClient stops a client
func (fe *FullEnv) StopClient(name string) error {
	if _, ok := fe.inClients[name]; ok {
		return fe.stopInProcessClient(name)
	} else if _, ok := fe.extClients[name]; ok {
		return fe.stopExtProcessClient(name)
	} else {
		return fmt.Errorf("No such client: %s", name)
	}
}

// Startup starts the clients
func (fe *FullEnv) Startup() error {
	for name := range fe.inClients {
		fe.StartClient(name)
	}
	for name := range fe.extClients {
		fe.StartClient(name)
	}
	return nil
}

// Shutdown stops the clients
func (fe *FullEnv) Shutdown() error {
	for name, cmd := range fe.extClients {
		if cmd != nil {
			fe.stopExtProcessClient(name)
		}
	}
	for name, running := range fe.runningInClients {
		if running {
			fe.stopInProcessClient(name)
		}
	}
	return nil
}

func captureCmdOutputs(name string, cmd *exec.Cmd, log xbus.Logger) error {
	stdout, err := cmd.StdoutPipe()
	if err != nil {
		return err
	}

	stderr, err := cmd.StderrPipe()
	if err != nil {
		return err
	}

	go func() {
		scanner := bufio.NewScanner(stdout)
		for scanner.Scan() {
			log.Noticef("-- " + name + ":out -- " + scanner.Text())
		}
	}()

	go func() {
		scanner := bufio.NewScanner(stderr)
		for scanner.Scan() {
			log.Noticef("-- " + name + ":err -- " + scanner.Text())
		}
	}()
	return nil
}

// StartXbusd starts the xbusd server
func (fe *FullEnv) StartXbusd() error {
	xbusdCfgFilename := path.Join(fe.dirname, "xbusd.yaml")
	p := os.Getenv("GOPATH")
	xbusd := exec.Command(
		path.Join(p, "bin", "xbusd"),
		"--config", xbusdCfgFilename,
		"serve")

	xbusd.Dir = fe.dirname

	if err := captureCmdOutputs("xbusd", xbusd, fe.log); err != nil {
		return err
	}

	if err := xbusd.Start(); err != nil {
		return err
	}

	fe.xbusd = xbusd
	fe.log.Noticef("xbusd started")
	time.Sleep(250 * time.Millisecond)
	return nil
}

// StopXbusd stops the xbusd server
func (fe *FullEnv) StopXbusd() error {
	if fe.xbusd != nil {
		fe.xbusd.Process.Signal(os.Interrupt)
		var timer *time.Timer
		timer = time.AfterFunc(3*time.Second, func() {
			fe.log.Errorf("Xbusd still running (deadlocked ?). Attempting a SIGQUIT")
			if err := fe.xbusd.Process.Signal(syscall.SIGQUIT); err != nil {
				fe.log.Errorf("SIGQUIT failed with '%s', killing the process.", err)
				fe.xbusd.Process.Kill()
			} else {
				timer = time.AfterFunc(time.Second, func() {
					fe.log.Errorf("Xbusd still running. Killing it")
					fe.xbusd.Process.Kill()
				})
			}
		})

		defer timer.Stop()
		if err := fe.xbusd.Wait(); err != nil {
			return err
		}
		fe.xbusd = nil
	}
	return nil
}

// Close stops everything, including the server
func (fe *FullEnv) Close() error {
	fe.Shutdown()
	fe.StopXbusd()
	err := fe.dblock.Unlock()
	if err != nil {
		fe.log.Errorf("Error releasing lock: %s", err)
		return err
	}
	for _, name := range fe.actorServices {
		xbus.UnregisterActorService(name)
	}
	fe.actorServices = nil
	if fe.t != nil && fe.t.Failed() {
		fe.t.Logf("Test failed. All the files can be found in %s", fe.dirname)
	}
	return nil
}

// CmdRunner can run a command with args
type CmdRunner struct {
	Dir     string
	Command string
	Args    []string
	Log     xbus.Logger
}

// NewCmdRunner initializes a CmdRunner
func NewCmdRunner(log xbus.Logger, dir, command string, args ...string) CmdRunner {
	return CmdRunner{dir, command, args, log}
}

// Run runs the command with args
func (r CmdRunner) Run(args ...string) error {
	cmd := exec.Command(r.Command, append(r.Args[:], args...)...)
	cmd.Dir = r.Dir

	b, err := cmd.CombinedOutput()
	if err != nil {
		r.Log.Errorf("'%s': %s", strings.Join(args, " "), b)
		return fmt.Errorf("Could not run '%s': %s", strings.Join(args, " "), err)
	}

	return nil
}

// RunReadLastLine runs the command with args and returns the last line of
// the command combined output
func (r CmdRunner) RunReadLastLine(args ...string) (string, error) {
	cmd := exec.Command(r.Command, append(r.Args[:], args...)...)
	cmd.Dir = r.Dir

	b, err := cmd.CombinedOutput()
	if err != nil {
		r.Log.Errorf("'%s': %s", strings.Join(args, " "), b)
		return "", fmt.Errorf("Could not run '%s': %s", strings.Join(args, " "), err)
	}

	lines := bytes.Split(b, []byte("\n"))
	var lastLine string
	for i := len(lines) - 1; i >= 0; i-- {
		if len(lines[i]) != 0 {
			lastLine = string(lines[i])
			break
		}
	}
	return lastLine, nil
}

// MustRun runs the command with args and panic if it fails
func (r CmdRunner) MustRun(args ...string) {
	if err := r.Run(args...); err != nil {
		panic(err)
	}
}

func must(err error, msg ...string) {
	if err != nil {
		panic(fmt.Errorf("%s: %s", strings.Join(msg, " "), err))
	}
}

func catcher(err *error) {
	if r := recover(); r != nil {
		if e, ok := r.(error); ok {
			*err = e
		} else {
			panic(r)
		}
	}
}
