package process

import (
	"fmt"

	"github.com/m4rw3r/uuid"

	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/xbus/graph"
	"bitbucket.org/orus-io/xbus/storage"
)

// EnvelopeActorPair contains a pair (EnvelopeID, ActorID)
type EnvelopeActorPair struct {
	EnvelopeID uuid.UUID
	ActorID    uuid.UUID
}

// NewProcess creates a process
func NewProcess(actorStorage storage.ActorStorage, emitterID, envelopeID uuid.UUID, graph graph.Graph, source storage.NodeOutputRef) *Process {
	id, err := uuid.V4()
	if err != nil {
		panic(fmt.Sprintf("Cannot generate a UUID: %s", err))
	}
	return &Process{
		ID:                id,
		actors:            actorStorage,
		SourceRef:         source,
		TriggerEmitterID:  emitterID,
		Graph:             graph,
		TriggerEnvelopeID: envelopeID,
	}
}

// Process orchestrate a graph execution
type Process struct {
	ID uuid.UUID

	actors         storage.ActorStorage
	processStorage storage.ProcessStorage

	TriggerEmitterID  uuid.UUID
	TriggerEnvelopeID uuid.UUID

	Graph     graph.Graph
	SourceRef storage.NodeOutputRef
}

// GetStatus returns the current status of the process
func (p Process) GetStatus() api.ProcessStatus {
	state := p.processStorage.GetState(p.ID)
	return state.Status
}

// Pause switch the process state to ProcessPaused
func (p Process) Pause() error {
	return p.processStorage.SetStatus(p.ID, api.ProcessPaused)
}

// Resume switch the process state to ProcessRunning if previously paused
func (p Process) Resume() error {
	return p.processStorage.SetStatus(p.ID, api.ProcessRunning)
}

// GetTargets returns the targets for an envelope coming from a given emitter
func (p Process) GetTargets(output storage.NodeOutputRef, envelopeID uuid.UUID, noRouteTableUpdate bool) ([]api.EnvelopeTarget, error) {
	// is the envelope already routed ?
	return p.processStorage.GetTargets(p.ID, output, envelopeID, noRouteTableUpdate,
		func() ([]api.EnvelopeTarget, error) {

			dests, err := p.Graph.GetDestinations(output)
			if err != nil {
				return nil, err
			}
			if len(dests) != 1 {
				fmt.Println(dests)
				panic("can only handle a single destination")
			}
			dest := dests[0]
			destNode, err := p.Graph.GetNode(dest.NodeID)
			if err != nil {
				return nil, fmt.Errorf("Cannot find the target node: %s", err)
			}

			var actors []api.Actor

			if destNode.Type == storage.ConsumerNode {
				if !destNode.RoleBroadcast || len(destNode.Roles) != 1 {
					panic("not implemented")
				}

				actors = p.actors.ListRole(api.ConsumerKind, destNode.Roles[0])
			} else if destNode.Type == storage.WorkerNode {
				if destNode.RoleBroadcast || len(destNode.Roles) != 0 {
					panic("Worker graph nodes supports only Actors, not roles")
				}

				for _, actorID := range destNode.ActorIDs {
					actor, exists := p.actors.Get(actorID)
					if !exists {
						return nil, fmt.Errorf("No such actor: %s", actorID)
					}
					actors = append(actors, actor)
				}

				for _, actorRef := range destNode.Actors {
					actorID := uuid.MaybeFromString(actorRef)
					if !actorID.IsZero() {
						actor, exists := p.actors.Get(actorID)
						if !exists {
							return nil, fmt.Errorf("No such actor: %s", actorID)
						}
						actors = append(actors, actor)
					} else {
						actor, exists := p.actors.GetByName(actorRef)
						if !exists {
							return nil, fmt.Errorf("No such actor: %s", actorRef)
						}
						actors = append(actors, actor)
					}
				}
			}

			if len(actors) == 0 {
				return nil, fmt.Errorf("No target actor found")
			}
			var targets []api.EnvelopeTarget
			for _, actor := range actors {
				targets = append(targets, api.EnvelopeTarget{
					ActorID: actor.ID,
					NodeID:  dest.NodeID,
					Input:   dest.Input,
				})
			}
			return targets, nil
		},
	)

}

// GetEmitPosition locks the emit position and returns an exclusive access to it
// It must be released by calling 'Unlock' on it
func (p *Process) GetEmitPosition(envelopeID uuid.UUID, target api.EnvelopeTarget) (storage.EmitPosition, error) {
	return p.processStorage.GetEmitPosition(storage.Emission{
		ProcessID:   p.ID,
		EnvelopeID:  envelopeID,
		TargetActor: target.ActorID,
		TargetInputRef: storage.NodeInputRef{
			NodeID: target.NodeID,
			Input:  target.Input,
		},
	})
}

// UpdateReceiveStatus is called whenever an envelope fragment is received from a node
func (p *Process) UpdateReceiveStatus(envelopeID, actorID uuid.UUID, outputref storage.NodeOutputRef, status storage.ReceptionStatus) error {
	return p.processStorage.UpdateReceiveStatus(
		storage.Reception{
			ProcessID:       p.ID,
			EnvelopeID:      envelopeID,
			SourceActor:     actorID,
			SourceOutputRef: outputref,
		},
		status)
}

// ActorProcessingEnd informs the process
func (p *Process) ActorProcessingEnd(envelopeID, actorID uuid.UUID, nodeID string, result api.ActorProcessingResultType) {
	p.processStorage.ProcessNodeEnd(
		p.ID, storage.ProcessNode{ActorID: actorID, NodeID: nodeID},
		envelopeID, result)
}

// GetPostMortemState set the postmortem state of the process
func (p *Process) GetPostMortemState() (storage.PostMortemState, error) {
	if status := p.GetStatus(); status != api.ProcessDone && status != api.ProcessError {
		return storage.PostMortemState{}, fmt.Errorf("Cannot get postmortem state on a non-ended process. Current status is '%s'", status)
	}
	return p.processStorage.GetPostMortemState(p.ID), nil
}

// SetPostMortemState set the postmortem state of the process
func (p *Process) SetPostMortemState(state storage.PostMortemState) error {
	if status := p.GetStatus(); status != api.ProcessDone && status != api.ProcessError {
		return fmt.Errorf("Cannot set postmortem state on a non-ended process. Current status is '%s'", status)
	}
	p.processStorage.SetPostMortemState(p.ID, state)
	return nil
}
