// Copyright 2017 Orus.io. All rights reserved.

package broker

import (
	"fmt"
	"strings"

	"github.com/m4rw3r/uuid"

	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/xbus/process"
	"bitbucket.org/orus-io/xbus/storage"
)

func parseSubject(subject string) (actorID uuid.UUID, output string, err error) {
	splitted := strings.Split(subject, ".")
	if len(splitted) != 5 {
		return uuid.UUID{}, "", fmt.Errorf("Invalid subject. Expected 5 parts, got %d", len(splitted))
	}
	if splitted[0] != "xbus" || splitted[1] != "msgbox" || splitted[3] != "output" {
		return uuid.UUID{}, "", fmt.Errorf("Invalid subject. Expected xbus.msgbox.*.output.*, got %s", subject)
	}
	actorID, err = uuid.FromString(splitted[2])
	if err != nil {
		return
	}
	output = splitted[4]
	return
}

func (s *Service) replyAckError(reply string, envelopeID uuid.UUID, err error) {
	s.Log.Errorf("error: %s", err)
	err = s.Conn.Publish(reply, api.NewEnvelopeAckError(envelopeID, err))
	if err != nil {
		s.Log.Errorf("Error replying: %s", err)
	}
	return
}

func (s *Service) receiveEnvelope(subject, reply string, envelope api.Envelope) (process *process.Process, sourceRef storage.NodeOutputRef, success bool) {
	s.Log.Debugf("Receiving envelope fragment")

	// Whatever happens, send a ACK
	var ack = api.EnvelopeAck{ID: envelope.ID}
	s.TaskAdd()
	defer s.SendReply(reply, &ack)

	// validate the subject matches a know actor
	actorID, output, err := parseSubject(subject)
	if err != nil {
		ack.Response = api.ClientError("Invalid subject '%s': %s", subject, err)
		return
	}

	actor, exists := s.Opts.ActorStorage.Get(actorID)
	if !exists {
		ack.Response = api.ClientError("Unknown actor: %s", actorID)
		return
	}

	// If no context is given, make sure the sender is as emitter
	if envelope.Context.ProcessID.IsZero() && actor.Kind != api.EmitterKind {
		msg := fmt.Sprintf("Context-less envelope sent by non-emitter %s", actor.Name)
		s.Log.Errorf(msg)
		ack.Response = api.ClientError(msg)
		return
	}

	// A context is given, make sure the process exists
	if !envelope.Context.ProcessID.IsZero() {
		process, err = s.processManager.Get(envelope.Context.ProcessID)
		if err != nil {
			err := fmt.Errorf("Invalid process on incoming envelope. Rejecting it")
			s.Log.Errorf("%s", err)
			err = s.Conn.Publish(reply, api.NewEnvelopeAckError(envelope.ID, err))
			return
		}
		// TODO make sure NodeID matches the sending actor
		sourceRef.NodeID = envelope.Context.NodeID
		sourceRef.Output = output
	}

	var (
		typesKnown bool
		eventTypes []string
	)

	// Check for empty envelope
	if envelope.Last == true && len(envelope.Events) == 0 && len(envelope.EventIDs) == 0 {
		if process == nil {
			s.replyAckError(
				reply, envelope.ID,
				fmt.Errorf("Invalid envelop (missing or invalid context)"))
			return
		}

		// empty envelope, which can mean 2 things depending on the ID

		// Without an ID: the envelope is just a signal that the output
		// will not be used in this context. Inform the process.
		if envelope.ID.IsZero() {
			ack.ReceptionStatus = api.ReceptionAccepted
			ack.Status = api.StatusOK

			process.UpdateReceiveStatus(
				envelope.ID, actor.ID, sourceRef, storage.ReceptionClosed)
			return
		}
		// With an ID: the envelope is one sent on an input, hence must be
		// known to the broker which can reread it from storage.
		ack.ReceptionStatus = api.ReceptionAccepted
		typesKnown = true
	} else {
		// non-empty envelope. Store it.
		status, err := s.Opts.EnvelopeStorage.StoreEnvelope(envelope)
		if err != nil {
			ack = api.NewEnvelopeAckError(envelope.ID, err)
			return
		}
		s.Log.Noticef("Stored envelope: %s", status.Status)
		switch status.Status {
		case "receiving":
			ack.ReceptionStatus = api.ReceptionReceiving
		case "complete":
			ack.ReceptionStatus = api.ReceptionAccepted
		case "error":
			ack.ReceptionStatus = api.ReceptionError
		default:
			ack.ReceptionStatus = api.ReceptionError
		}
		typesKnown = status.TypesKnown
		eventTypes = status.EventTypes()

		if ack.ReceptionStatus == api.ReceptionError {
			if process != nil {
				process.UpdateReceiveStatus(
					envelope.ID, actor.ID, sourceRef, storage.ReceptionError)
			}
			ack = api.NewEnvelopeAckError(envelope.ID, status)
			return
		}
	}

	// The envelope is now safely stored
	// We just need the process to be aware of this reception, and we will
	// be done.

	// No context but all types known -> we can get a process
	if envelope.Context.ProcessID.IsZero() && typesKnown {
		process, err = s.processManager.LookupOrCreate(actor, envelope.ID, eventTypes)
		if err != nil {
			ack.ReceptionStatus = api.ReceptionError
			ack.Reason = fmt.Sprintf(
				"Error getting process for incoming envelope %s: %s", envelope.ID, err)
			s.Log.Errorf(ack.Reason)
			return
		}
		s.Log.Debugf("Process ID: %s", process.ID)
		sourceRef = process.SourceRef
	}

	// If process is unknown and the reception is over, we have a serious
	// problem
	if process == nil && ack.ReceptionStatus == api.ReceptionAccepted {
		panic("Impossible state, please report")
	}

	if process != nil {
		// If a process is associated to the envelope, make sure the route for the next
		// move is known, then update the reception status.
		// Logically we should calculate the route after updating the status, but if the
		// server crashes we would have route-less receptions that would be hard to
		// track and fix. Switching the two operations is harmless: a reception-less
		// route will become used as soon as the emitter re-attempt its sending.
		// And because the route is present, the emission service will find what
		// to do as soon as it is (re)started
		// See issue #39

		_, err := process.GetTargets(sourceRef, envelope.ID, false)
		if err != nil {
			panic(fmt.Sprintf("Error getting routes: %s", err))
		}

		var storageReceptionStatus storage.ReceptionStatus
		switch ack.ReceptionStatus {
		case api.ReceptionReceiving:
			storageReceptionStatus = storage.ReceptionOngoing
		case api.ReceptionAccepted:
			storageReceptionStatus = storage.ReceptionDone
		}
		if err := process.UpdateReceiveStatus(
			envelope.ID, actor.ID, sourceRef, storageReceptionStatus,
		); err != nil {
			panic(fmt.Sprintf("Error updating receive status: %s", err))
		}
	}

	// our part (receive the envelope fragment) is done, we can safely
	// let the ack be sent to the sender
	ack.Response = api.Response{Status: api.StatusOK}

	return process, sourceRef, true
}

func (s *Service) handleIncomingEnvelope(subject, reply string, envelope api.Envelope) {
	s.TaskAdd()
	defer s.TaskDone()

	// XXX we do not want to block other incoming envelopes, we should have workers
	// doing the job in the background so we can handle a number of concurrent incoming
	// envelopes
	// We do not want to fire a route for each incoming envelope either,
	// to avoid handling too many envelopes at the same time

	// handle the reception
	process, sourceRef, success := s.receiveEnvelope(subject, reply, envelope)

	// to the routing if relevant
	if success && process != nil {
		s.routeEnvelope(process, envelope.ID, sourceRef)
	}
}

func (s *Service) routeEnvelope(
	process *process.Process,
	envelopeID uuid.UUID,
	sourceRef storage.NodeOutputRef,
) {
	targets, err := process.GetTargets(sourceRef, envelopeID, false)
	if err != nil {
		s.Log.Errorf("Error getting targets: %s", err)
		return
	}

	s.Log.Noticef("Will forward to %s", targets)

	for _, target := range targets {
		s.Conn.Publish(api.EmissionSignalSubjet, api.EmissionSignal{
			ProcessID:  process.ID,
			EnvelopeID: envelopeID,
			Target: api.EnvelopeTarget{
				ActorID: target.ActorID,
				NodeID:  target.NodeID,
				Input:   target.Input,
			},
		})
	}
}
