// Copyright 2017 Orus.io. All rights reserved.

package clientapi

import (
	"fmt"
	"strings"

	"github.com/m4rw3r/uuid"

	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/xbus/services"
	"bitbucket.org/orus-io/xbus/storage"
)

// Options for the Service
type Options struct {
	services.Options

	ActorStorage   storage.ActorStorage
	AccountStorage storage.AccountStorage
}

// NewService instanciate a Service
func NewService(o Options) *Service {
	return &Service{
		o:           o,
		BaseService: services.NewBaseService("clientapi", o.Options),
	}
}

// Service provides an api for administrators
type Service struct {
	services.BaseService

	o Options
}

// Startup starts the service
func (s *Service) Startup() error {
	whoamiSubject, _ := api.WhoAmISubjects("*")
	if err := s.SubscribeAll(
		services.Subscribe(
			api.ClientSubject(uuid.UUID{}, api.ClientGetActors),
			s.handleClientGetActors),
		services.Subscribe(
			whoamiSubject,
			s.handleWhoAmI),
	); err != nil {
		s.Shutdown()
		return err
	}
	return nil
}

func (s *Service) parseSubject(subject string) (uuid.UUID, string, error) {
	splitted := strings.Split(subject, ".")
	if len(splitted) != 4 {
		return uuid.UUID{}, "", fmt.Errorf("Invalid subject: %s (expected 4 parts)", subject)
	}
	if splitted[0] != "xbus" || splitted[1] != "client" {
		return uuid.UUID{}, "", fmt.Errorf("Invalid subject: %s (invalid prefix, expected 'xbus.client')", subject)
	}
	id, err := uuid.FromString(splitted[2])
	if err != nil {
		return uuid.UUID{}, "", fmt.Errorf("Invalid subject: %s (invalid uuid in 3 position)", subject)
	}
	return id, splitted[3], nil
}

func (s *Service) handleClientGetActors(subject, reply string, req api.ClientGetActorsReq) {
	var rep api.ClientGetActorsRep
	s.TaskAdd()
	defer s.SendReply(reply, &rep)

	accountID, fname, err := s.parseSubject(subject)
	if err != nil {
		panic(err)
	}
	if fname != "get_actors" {
		panic("Invalid function name: " + fname)
	}

	rep.Actors = s.o.ActorStorage.ListAccount(accountID)
	rep.Status = api.StatusOK
}

func (s *Service) handleWhoAmI(subject, reply string, req string) {
	var rep api.WhoAmIRep
	s.TaskAdd()
	defer s.SendReply(reply, &rep)

	splitted := strings.Split(subject, ".")

	apiKey := splitted[3]

	account, ok := s.o.AccountStorage.GetByAPIKey(apiKey)
	if !ok {
		panic("No account by this api key")
	}
	rep.Status = api.StatusOK
	rep.Account = account
}
