package control

import (
	"crypto/x509"
	"encoding/pem"
	"fmt"
	"math/rand"
	"time"

	"github.com/m4rw3r/uuid"

	"bitbucket.org/orus-io/go-xbus"
	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/go-xbus/tools"
	"bitbucket.org/orus-io/xbus/storage"
)

var letters = []rune("0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ")

func randSeq(n int) string {
	b := make([]rune, n)
	for i := range b {
		b[i] = letters[rand.Intn(len(letters))]
	}
	return string(b)
}

// AccountAccept accepts a pending account request; a certificate is generated from the CSR.
func AccountAccept(
	accountStorage storage.AccountStorage, accountID uuid.UUID,
	serverCert *x509.Certificate, serverKey interface{},
) (api.Account, error) {
	a, exists := accountStorage.Get(accountID)
	if !exists {
		return api.Account{}, fmt.Errorf("No such account: %s", accountID)
	}
	if a.Status != api.AccountStatusPending {
		return api.Account{}, fmt.Errorf("Invalid initial status %s for 'Accept'", a.Status)
	}
	p, r := pem.Decode([]byte(a.CSR))
	if len(r) != 0 {
		return api.Account{}, fmt.Errorf("Invalid CSR pem encoding")
	}
	if p.Type != "CERTIFICATE REQUEST" {
		return api.Account{}, fmt.Errorf("Invalid CSR pem encoding")
	}
	cert, err := tools.CreateCertificateFromCSR(
		p.Bytes, serverCert, serverKey, time.Hour*24*365*2,
	)
	if err != nil {
		return api.Account{}, err
	}
	a.Cert = string(pem.EncodeToMemory(&pem.Block{
		Type:  "CERTIFICATE",
		Bytes: cert,
	}))
	a.Status = api.AccountStatusEnabled
	return accountStorage.Update(a)
}

// AccountReject rejects a pending account request.
func AccountReject(
	accountStorage storage.AccountStorage, accountID uuid.UUID,
) (api.Account, error) {
	a, exists := accountStorage.Get(accountID)
	if !exists {
		return api.Account{}, fmt.Errorf("No such account: %s", accountID)
	}
	if a.Status != api.AccountStatusPending {
		return api.Account{}, fmt.Errorf("Invalid initial status %q for 'Reject'", a.Status)
	}
	a.Status = api.AccountStatusDisabled
	return accountStorage.Update(a)
}

func accountDelete(storage storage.AccountStorage, accountID uuid.UUID) error {
	a, exists := storage.Get(accountID)
	if !exists {
		return xbus.ErrNotFound
	}
	storage.Delete(a)
	return nil
}

// handleAccountDelete is the event handler called when a request is received on the
// "api.ControlAccountDelete" subject.
func (s *Service) handleAccountDelete(subject, reply string, req api.ControlAccountDeleteReq) {
	var rep api.ControlAccountDeleteRep
	var err error

	s.TaskAdd()
	defer s.SendReply(reply, &rep)

	err = accountDelete(s.o.AccountStorage, uuid.UUID(req))
	if err != nil {
		if err == xbus.ErrNotFound {
			rep.Response = api.ClientError("No such account")
		}
		rep.Response = api.ServerError(err)
	}
	rep.Status = api.StatusOK
}

// handleAccountList is the event handler called when a request is received on the
// "api.ControlAccountList" subject.
func (s *Service) handleAccountList(subject, reply string, req api.ControlAccountListReq) {
	var rep api.ControlAccountListRep
	s.TaskAdd()
	defer s.SendReply(reply, &rep)

	rep.AccountList = s.o.AccountStorage.List()
	rep.Status = api.StatusOK
}

// handleAccountUpdate is the event handler called when a request is received on the
// "api.ControlAccountUpdate" subject.
func (s *Service) handleAccountUpdate(subject, reply string, req api.ControlAccountUpdateReq) {
	var rep api.ControlAccountUpdateRep
	s.TaskAdd()
	defer s.SendReply(reply, &rep)

	if req.Account.ID.IsZero() {
		rep.Response = api.ClientError("Must pass an ID")
		return
	}
	a, exists := s.o.AccountStorage.Get(req.Account.ID)
	if !exists {
		rep.Response = api.ClientError("No such account: %s", req.Account.ID)
		return
	}
	var err error
	if req.Account.Status == api.AccountStatusEnabled && a.Status == api.AccountStatusPending {
		a, err = AccountAccept(s.o.AccountStorage, req.Account.ID, s.o.ServerCert, s.o.ServerKey)
	} else if req.Account.Status == api.AccountStatusDisabled && a.Status == api.AccountStatusPending {
		a, err = AccountReject(s.o.AccountStorage, req.Account.ID)
	} else {
		rep.Response = api.ClientError("Invalid change")
		return
	}
	if err != nil {
		rep.Response = api.ClientError(err.Error())
		return
	}
	rep.Account = api.Account{
		ID:     a.ID,
		Name:   a.Name,
		Type:   a.Type,
		Status: a.Status,
		CSR:    a.CSR,
		Cert:   a.Cert,
	}
	rep.Status = api.StatusOK
}

// handleAccountRenewAPIKey
func (s *Service) handleAccountRenewAPIKey(subject, reply string, req api.ControlAccountRenewAPIKeyReq) {
	var rep api.ControlAccountRenewAPIKeyRep
	s.TaskAdd()
	defer s.SendReply(reply, &rep)

	account, ok := s.o.AccountStorage.Get(uuid.UUID(req))
	if !ok {
		rep.Response = api.ClientError("No such account")
		return
	}

	account.APIKey = randSeq(32)
	account, err := s.o.AccountStorage.Update(account)
	if err != nil {
		rep.Response = api.ServerError(err)
		return
	}
	rep.Account = account
	rep.Status = api.StatusOK
}
