// Copyright 2016 Orus.io. All rights reserved.

package sql

import (
	"encoding/json"
	"fmt"
	"strconv"
	"time"

	"github.com/m4rw3r/uuid"
	"github.com/orus-io/yago"
	"github.com/slicebit/qb"

	"bitbucket.org/orus-io/go-xbus"
	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/xbus/storage"
	"bitbucket.org/orus-io/xbus/storage/sql/model"
)

// NewProcessStorage initialize a new ProcessStorage
func NewProcessStorage(model *model.Model, db *yago.DB, logger xbus.Logger) *ProcessStorage {
	return &ProcessStorage{
		Base: Base{
			model:  model,
			db:     db,
			logger: logger,
		},
		nextState:       storage.UndefinedNextStateCallback,
		onStatusChanged: storage.NoOpProcessEventStatusHandler,
	}
}

// ProcessStorage implements storage.ProcessStorage
type ProcessStorage struct {
	Base
	nextState       storage.NextStateCallback
	onStatusChanged storage.ProcessEventStatusHandler
}

// SetEventHandlers sets the event handlers
func (s *ProcessStorage) SetEventHandlers(onStatusChanged storage.ProcessEventStatusHandler) {
	if onStatusChanged == nil {
		s.onStatusChanged = storage.NoOpProcessEventStatusHandler
	} else {
		s.onStatusChanged = onStatusChanged
	}
}

func dbProcessToStorage(dbProcess model.Process) storage.Process {
	return storage.Process{
		ID:                dbProcess.ID,
		TriggerEmitterID:  dbProcess.TriggerEmitterID,
		TriggerEnvelopeID: dbProcess.TriggerEnvelopeID,
		GraphID:           dbProcess.GraphID,
		SourceOutput: storage.NodeOutputRef{
			NodeID: dbProcess.SourceNodeID,
			Output: dbProcess.SourceSlotID,
		},
	}
}

func storageProcessToDB(process storage.Process) model.Process {
	return model.Process{
		Base:              model.Base{ID: process.ID},
		TriggerEmitterID:  process.TriggerEmitterID,
		TriggerEnvelopeID: process.TriggerEnvelopeID,
		GraphID:           process.GraphID,
		SourceNodeID:      process.SourceOutput.NodeID,
		SourceSlotID:      process.SourceOutput.Output,
	}
}

func (s *ProcessStorage) loadProcess(tx *yago.Tx, id uuid.UUID) (storage.Process, bool) {
	var dbProcess model.Process
	if err := tx.Query(s.model.Process).Get(&dbProcess, id); err != nil {
		if err == yago.ErrRecordNotFound {
			return storage.Process{}, false
		}
		must(err)
	}
	return dbProcessToStorage(dbProcess), true
}

// GetProcess returns a process data
func (s *ProcessStorage) GetProcess(id uuid.UUID) (storage.Process, bool) {
	tx := s.MustBegin()
	defer tx.Rollback()
	return s.loadProcess(tx.Tx, id)
}

// GetProcessByTrigger returns a process matching a trigger or an empty value if not found
// TODO in the future we should support multiple process per trigger.
func (s *ProcessStorage) GetProcessByTrigger(emitter api.Actor, envelopeID uuid.UUID) (storage.Process, bool) {
	tx := s.MustBegin()
	defer tx.Rollback()
	var dbProcess model.Process
	if err := tx.Query(s.model.Process).Where(
		s.model.Process.TriggerEmitterID.Eq(emitter.ID),
		s.model.Process.TriggerEnvelopeID.Eq(envelopeID),
	).One(&dbProcess); err != nil {
		if err == yago.ErrRecordNotFound {
			return storage.Process{}, false
		}
		must(err)
	}
	return dbProcessToStorage(dbProcess), true
}

// Query lookup for processes
func (s *ProcessStorage) Query(includeTerminated bool) []storage.Process {
	var dbProcessList []model.Process
	tx := s.MustBegin()
	defer tx.Rollback()
	q := tx.Query(s.model.Process)
	if !includeTerminated {
		q = q.LeftJoin(s.model.ProcessState).Filter(
			s.model.ProcessState.Status.Eq(api.ProcessRunning),
		)
	}
	must(
		q.All(&dbProcessList))
	var pList []storage.Process
	for _, dbP := range dbProcessList {
		pList = append(pList, dbProcessToStorage(dbP))
	}
	return pList
}

// CreateProcess saves a process
func (s *ProcessStorage) CreateProcess(p storage.Process) (uuid.UUID, error) {
	tx := s.MustBegin()
	defer tx.RollbackIfOpened()

	p.ID = uuid.UUID{}

	dbProcess := storageProcessToDB(p)

	must(tx.Insert(&dbProcess))

	tx.MustCommit()

	return dbProcess.ID, nil
}

// GetTargets returns the targets for an envelope if already known, or 'nil' if not
func (s *ProcessStorage) GetTargets(
	processID uuid.UUID,
	source storage.NodeOutputRef,
	envelopeID uuid.UUID,
	noRouteTableUpdate bool,
	calcTargets func() ([]api.EnvelopeTarget, error),
) (targets []api.EnvelopeTarget, calcErr error) {
	tx := s.MustBegin()
	defer tx.RollbackIfOpened()

	bSource, err := json.Marshal(source)
	if err != nil {
		panic(err)
	}

	Route, RouteTarget := s.model.Route, s.model.RouteTarget

	var dbTargetList []model.RouteTarget

	if err := tx.Query(RouteTarget).
		LeftJoin(Route).
		Where(
			Route.ProcessID.Eq(processID),
			Route.Source.Eq(bSource),
			Route.EnvelopeID.Eq(envelopeID)).
		All(&dbTargetList); err != nil {
		panic(err)
	}

	if len(dbTargetList) != 0 {
		for _, dbTarget := range dbTargetList {
			targets = append(targets, api.EnvelopeTarget{
				ActorID: dbTarget.ActorID,
				NodeID:  dbTarget.NodeID,
				Input:   dbTarget.Input,
			})
		}
		return
	}

	tx.Rollback()

	targets, calcErr = calcTargets()
	if calcErr != nil || noRouteTableUpdate {
		return
	}
	route := model.Route{
		ProcessID:  processID,
		Source:     string(bSource),
		EnvelopeID: envelopeID,
	}

	tx2 := s.MustBegin()
	defer tx2.RollbackIfOpened()

	// Use an exclusive transaction
	// on PG: SET TRANSACTION SERIALIZABLE
	if false && s.db.Engine.Dialect().Driver() == "postgres" {
		_, err := tx2.GetEngine().Exec(model.RawSQL("SET TRANSACTION ISOLATION LEVEL SERIALIZABLE"))
		if err != nil {
			panic(err)
		}
	}

	if err := tx2.Query(RouteTarget).
		LeftJoin(Route).
		Where(
			Route.ProcessID.Eq(processID),
			Route.Source.Eq(bSource),
			Route.EnvelopeID.Eq(envelopeID)).
		All(&dbTargetList); err != nil {
		panic(err)
	}

	// The routes where calculated faster by someone else
	// We return their result and ignore ours
	if len(dbTargetList) != 0 {
		for _, dbTarget := range dbTargetList {
			targets = append(targets, api.EnvelopeTarget{
				ActorID: dbTarget.ActorID,
				NodeID:  dbTarget.NodeID,
				Input:   dbTarget.Input,
			})
		}
		return
	}

	var logentries []storage.ProcessLogEntry
	// TODO: detect a concurrent insert and handle it gently
	must(tx2.Insert(&route))
	for _, target := range targets {
		must(tx2.Insert(&model.RouteTarget{
			RouteID: route.ID,
			ActorID: target.ActorID,
			NodeID:  target.NodeID,
			Input:   target.Input,
		}))
		logentry, err := s.processLogEmission(
			tx2,
			processID,
			storage.Emission{
				ProcessID:   processID,
				EnvelopeID:  envelopeID,
				TargetActor: target.ActorID,
				TargetInputRef: storage.NodeInputRef{
					NodeID: target.NodeID,
					Input:  target.Input,
				},
			},
			storage.EnvelopeStartPos())
		must(err)
		logentries = append(logentries, logentry)
	}
	s.logger.Noticef("New routes -> logentries = %+v", logentries)
	must(s.updateStateAndCommit(tx2, processID, logentries))
	return

}

// UpdateReceiveStatus updates a reception status
func (s *ProcessStorage) UpdateReceiveStatus(reception storage.Reception, status storage.ReceptionStatus) error {
	tx := s.MustBegin()
	defer tx.RollbackIfOpened()

	logentry, err := s.processLogReception(tx, reception.ProcessID, reception, status)
	must(err)

	return s.updateStateAndCommit(
		tx, reception.ProcessID, []storage.ProcessLogEntry{logentry})
}

// ProcessNodeEnd signals the end of a ProcessNode.
func (s *ProcessStorage) ProcessNodeEnd(
	processID uuid.UUID, node storage.ProcessNode,
	envelopeID uuid.UUID, result api.ActorProcessingResultType,
) error {
	tx := s.MustBegin()
	defer tx.RollbackIfOpened()

	logentry, err := s.processLogProcessNodeEnd(
		tx, processID,
		storage.ProcessNodeEnd{
			Node:       node,
			EnvelopeID: envelopeID,
			Result:     result,
		})
	must(err)

	return s.updateStateAndCommit(
		tx, processID, []storage.ProcessLogEntry{logentry})
}

// GetEmitPosition returns an exclusive access to an emit position
func (s *ProcessStorage) GetEmitPosition(emission storage.Emission) (storage.EmitPosition, error) {
	var ownTx = true
	tx := s.MustBegin()
	defer func() {
		if ownTx {
			tx.Rollback()
		}
	}()
	// TODO Use an exclusive transaction
	// on PG: SET TRANSACTION SERIALIZABLE
	if false && s.db.Engine.Dialect().Driver() == "postgres" {
		_, err := tx.GetEngine().Exec(model.RawSQL("SET TRANSACTION ISOLATION LEVEL SERIALIZABLE"))
		if err != nil {
			panic(err)
		}
	}
	// They load the value from the db
	var dbEmission model.Emission
	var pos = storage.EnvelopeStartPos()
	q := tx.Query(s.model.Emission).
		Where(
			s.model.Emission.ProcessID.Eq(emission.ProcessID),
			s.model.Emission.EnvelopeID.Eq(emission.EnvelopeID),
			s.model.Emission.TargetActor.Eq(emission.TargetActor),
			s.model.Emission.TargetNodeID.Eq(emission.TargetInputRef.NodeID),
			s.model.Emission.TargetInput.Eq(emission.TargetInputRef.Input),
		).
		ForUpdate(s.model.Emission)
	if err := q.One(&dbEmission); err != nil {
		if err == yago.ErrRecordNotFound {
			// XXX The following code is pg specific
			_, insertErr := tx.GetEngine().Exec(model.RawSQL(
				"INSERT INTO emission "+
					"(process_id, envelope_id, target_actor, target_node_id, "+
					"target_input, start, complete, event_positions)"+
					" VALUES ($1, $2, $3, $4, $5, $6, $7, $8) "+
					" ON CONFLICT DO NOTHING",
				emission.ProcessID,
				emission.EnvelopeID,
				emission.TargetActor,
				emission.TargetInputRef.NodeID,
				emission.TargetInputRef.Input,
				true,
				false,
				"{}",
			))
			must(insertErr)
			// Retrieve the newly inserted emission or the one that was inserted
			// just before we did (in which case the insert should do nothing
			err = q.One(&dbEmission)
		}
		if err != nil {
			panic(err)
		}
	}
	pos.Start = dbEmission.Start
	pos.Complete = dbEmission.Complete
	if err := json.Unmarshal([]byte(dbEmission.EventPositions), &pos.Events); err != nil {
		panic(err)
	}
	ownTx = false
	return &EmitPosition{
		tx:             tx,
		emission:       dbEmission,
		pos:            pos,
		processStorage: s,
	}, nil
}

func (s *ProcessStorage) list(emissionClauses []qb.Clause, routeClauses []qb.Clause) []storage.Emission {
	Emission, Route, RouteTarget := s.model.Emission, s.model.Route, s.model.RouteTarget
	var dbEmissionList []model.Emission
	tx := s.MustBegin()
	defer tx.Rollback()
	must(
		tx.Query(Emission).
			Where(emissionClauses...).
			All(&dbEmissionList))
	var emissionList []storage.Emission
	for _, e := range dbEmissionList {
		emissionList = append(emissionList, storage.Emission{
			ProcessID:   e.ProcessID,
			EnvelopeID:  e.EnvelopeID,
			TargetActor: e.TargetActor,
			TargetInputRef: storage.NodeInputRef{
				NodeID: e.TargetNodeID,
				Input:  e.TargetInput,
			},
		})
	}

	q := tx.Query(RouteTarget).
		LeftJoin(Route).
		Where(routeClauses...).
		Filter(
			qb.NotExists(
				tx.Query(Emission).Where(
					Emission.ProcessID.Eq(Route.ProcessID),
					Emission.EnvelopeID.Eq(Route.EnvelopeID),
					Emission.TargetActor.Eq(RouteTarget.ActorID),
					Emission.TargetNodeID.Eq(RouteTarget.NodeID),
					Emission.TargetInput.Eq(RouteTarget.Input)).SelectStmt())).
		Select(
			Route.ProcessID,
			Route.EnvelopeID,
			RouteTarget.ActorID,
			RouteTarget.NodeID,
			RouteTarget.Input)

	stmt := q.SelectStmt().Build(s.db.Engine.Dialect())
	s.logger.Noticef("Listing unstarted routes: %s - %s",
		stmt.SQL(), stmt.Bindings(),
	)

	rows, err := q.SQLQuery()
	must(err)
	defer rows.Close()
	for rows.Next() {
		var emission storage.Emission
		must(rows.Scan(&emission.ProcessID,
			&emission.EnvelopeID,
			&emission.TargetActor,
			&emission.TargetInputRef.NodeID,
			&emission.TargetInputRef.Input))
		s.logger.Noticef("Adding dummy emission from unstarted route: %+v", emission)
		emissionList = append(emissionList, emission)
	}
	return emissionList
}

// ListActiveEmissions returns the active emissions
func (s *ProcessStorage) ListActiveEmissions(targetActor uuid.UUID) []storage.Emission {
	return s.list(
		[]qb.Clause{
			s.model.Emission.Complete.Eq(false),
			s.model.Emission.TargetActor.Eq(targetActor)},
		[]qb.Clause{
			s.model.RouteTarget.ActorID.Eq(targetActor),
		},
	)
}

// ListProcessEmissions returns the active emissions
func (s *ProcessStorage) ListProcessEmissions(process uuid.UUID) []storage.Emission {
	return s.list(
		[]qb.Clause{
			s.model.Emission.Complete.Eq(false),
			s.model.Emission.ProcessID.Eq(process)},
		[]qb.Clause{
			s.model.Route.ProcessID.Eq(process),
		},
	)
}

// GetState returns an instant copy of the state of a process
func (s *ProcessStorage) GetState(processID uuid.UUID) storage.ProcessState {
	tx := s.MustBegin()
	defer tx.Rollback()
	state, err := s.loadState(tx.Tx, processID, false)
	must(err)
	return state
}

// SetStatus changes the current status. Only Runnning->Paused and Paused->Running
// are allowed at this point
func (s *ProcessStorage) SetStatus(processID uuid.UUID, status api.ProcessStatus) error {
	tx := s.MustBegin()
	defer tx.RollbackIfOpened()

	process, ok := s.loadProcess(tx.Tx, processID)
	if !ok {
		return fmt.Errorf("No such process: %s", processID)
	}

	state, err := s.loadState(tx.Tx, processID, false)
	must(err)

	if !(status == api.ProcessPaused && state.Status == api.ProcessRunning ||
		status == api.ProcessRunning && state.Status == api.ProcessPaused) {
		return fmt.Errorf("Invalid state transition: %s -> %s", state.Status, status)
	}

	old := state.Status
	state.Status = status

	err = s.saveState(tx.Tx, processID, state)
	must(err)

	tx.MustCommit()

	s.onStatusChanged(process, old, state.Status)

	return nil
}

// SetNextStateCallback ...
func (s *ProcessStorage) SetNextStateCallback(cb storage.NextStateCallback) {
	if cb == nil {
		s.nextState = storage.UndefinedNextStateCallback
	} else {
		s.nextState = cb

		// Check for pending ProcessLogEntry that remains unprocessed because
		// of a server crash.
		// XXX It only works as long as we have only one processStorage running
		// When we move to distributed, we need to refactor this.
		s.handlePendingProcessLogEntry()
	}
}

func (s *ProcessStorage) loadState(tx *yago.Tx, processID uuid.UUID, forUpdate bool) (pState storage.ProcessState, err error) {
	var dbProcessState model.ProcessState

	q := tx.Query(s.model.ProcessState)
	if forUpdate {
		q = q.ForUpdate()
	}
	if err = q.Get(&dbProcessState, processID); err != nil && err != yago.ErrRecordNotFound {
		return pState, err
	}

	if dbProcessState.NodeStates != "" {
		pState.Status = dbProcessState.Status
		if err = json.Unmarshal(
			[]byte(dbProcessState.NodeStates),
			&pState.NodeStates); err != nil {
			return
		}
	}
	return pState, nil
}

func (s *ProcessStorage) saveState(tx *yago.Tx, processID uuid.UUID, pState storage.ProcessState) error {
	b, err := json.Marshal(pState.NodeStates)
	if err != nil {
		return err
	}
	dbProcessState := model.ProcessState{
		ProcessID:  processID,
		Status:     pState.Status,
		NodeStates: string(b),
	}
	exists, err := tx.Query(s.model.ProcessState).
		Where(s.model.ProcessState.ProcessID.Eq(processID)).
		Exists()
	if err != nil {
		return err
	}
	if !exists {
		return tx.Insert(&dbProcessState)
	}
	return tx.Update(&dbProcessState)
}

func (s *ProcessStorage) processLogEmission(tx *TxHelper, processID uuid.UUID, emission storage.Emission, pos storage.EnvelopeReadPos) (storage.ProcessLogEntry, error) {
	logentry := storage.ProcessLogEntry{
		Emission:         &emission,
		EmissionPosition: &pos,
	}
	return s.processLog(tx, processID, logentry)
}

func (s *ProcessStorage) processLogReception(tx *TxHelper, processID uuid.UUID, reception storage.Reception, status storage.ReceptionStatus) (storage.ProcessLogEntry, error) {
	logentry := storage.ProcessLogEntry{
		Reception:       &reception,
		ReceptionStatus: status,
	}
	return s.processLog(tx, processID, logentry)
}

func (s *ProcessStorage) processLogProcessNodeEnd(
	tx *TxHelper, processID uuid.UUID, nodeEnd storage.ProcessNodeEnd,
) (storage.ProcessLogEntry, error) {
	return s.processLog(
		tx, processID,
		storage.ProcessLogEntry{
			ProcessNodeEnd: &nodeEnd,
		})
}

func (s *ProcessStorage) processLog(tx *TxHelper, processID uuid.UUID, logentry storage.ProcessLogEntry) (storage.ProcessLogEntry, error) {
	logentry.Timestamp = time.Now()
	b, err := json.Marshal(logentry)
	if err != nil {
		return storage.ProcessLogEntry{}, err
	}
	dbLogEntry := model.ProcessLogEntry{
		ProcessID: processID,
		Timestamp: logentry.Timestamp,
		Data:      string(b),
	}
	if err := tx.Insert(&dbLogEntry); err != nil {
		return storage.ProcessLogEntry{}, err
	}
	logentry.ID = strconv.FormatInt(dbLogEntry.ID, 10)
	return logentry, nil
}

func (s *ProcessStorage) updateState(processID uuid.UUID, logentries []storage.ProcessLogEntry) error {
	process, ok := s.GetProcess(processID)
	if !ok {
		return fmt.Errorf("No such process:  %s", processID)
	}
	// Calculate the next state
	pState := s.GetState(processID)
	statusList := []api.ProcessStatus{pState.Status}
	for _, logentry := range logentries {
		var err error
		pState, err = s.nextState(process, pState, logentry)
		must(err)

		if pState.Status != statusList[len(statusList)-1] {
			statusList = append(statusList, pState.Status)
		}
	}

	// Mark the entries and save the new state
	tx := s.MustBegin()
	defer tx.RollbackIfOpened()

	for _, logentry := range logentries {
		id, err := strconv.ParseInt(logentry.ID, 10, 64)
		must(err)
		must(tx.Update(
			&model.ProcessLogEntry{ID: id, LogProcessed: true},
			"LogProcessed"))
	}

	must(
		s.saveState(tx.Tx, processID, pState))

	tx.MustCommit()

	// Emit a StatusChanged for each transition
	for i := range statusList[1:] {
		s.onStatusChanged(process, statusList[i], statusList[i+1])
	}
	return nil
}

func (s *ProcessStorage) updateStateAndCommit(tx *TxHelper, processID uuid.UUID, logentries []storage.ProcessLogEntry) error {
	// Commit the transaction so nextState is called outside it
	tx.MustCommit()

	return s.updateState(processID, logentries)
}

func (s *ProcessStorage) handlePendingProcessLogEntry() error {
	tx := s.MustBegin()
	defer tx.RollbackIfOpened()

	var dbLogEntries []model.ProcessLogEntry

	q := tx.Query(s.model.ProcessLogEntry).
		Where(s.model.ProcessLogEntry.LogProcessed.Eq(false))
	must(q.All(&dbLogEntries))

	must(tx.Rollback())

	for _, dbLogEntry := range dbLogEntries {
		var logEntry storage.ProcessLogEntry
		if err := json.Unmarshal([]byte(dbLogEntry.Data), &logEntry); err != nil {
			panic(err)
		}
		logEntry.ID = strconv.FormatInt(dbLogEntry.ID, 10)
		if err := s.updateState(dbLogEntry.ProcessID, []storage.ProcessLogEntry{logEntry}); err != nil {
			panic(err)
		}
	}

	return nil
}

// QueryPostMortem returns a list of dead processes
func (s *ProcessStorage) QueryPostMortem(level api.LogLevel, statusList ...api.PostMortemStatus) []storage.Process {
	tx := s.MustBegin()
	defer tx.Rollback()

	q := tx.Query(s.model.Process).
		LeftJoin(s.model.ProcessState)
	if len(statusList) != 0 {
		var crits []qb.Clause
		for _, status := range statusList {
			crits = append(crits, s.model.ProcessState.PMStatus.Eq(status))
		}
		q = q.Filter(qb.Or(crits...))
	}
	if level == api.LogError {
		q = q.Filter(s.model.ProcessState.PMLevel.Eq(api.LogError))
	}

	var dbProcessList []model.Process
	must(
		q.All(&dbProcessList))
	var pList []storage.Process
	for _, dbP := range dbProcessList {
		pList = append(pList, dbProcessToStorage(dbP))
	}
	return pList
}

// GetPostMortemState returns the postmortem state of a process
func (s *ProcessStorage) GetPostMortemState(id uuid.UUID) storage.PostMortemState {
	tx := s.MustBegin()
	defer tx.Rollback()

	row := tx.Query(s.model.ProcessState).
		Where(s.model.ProcessState.ProcessID.Eq(id)).
		Select(
			s.model.ProcessState.PMLevel,
			s.model.ProcessState.PMStatus,
			s.model.ProcessState.PMComment).
		SQLQueryRow()

	var pmLevel, pmStatus []byte
	var pmState storage.PostMortemState

	must(row.Scan(&pmLevel, &pmStatus, &pmState.Comment))
	must(pmState.Level.UnmarshalText(pmLevel))
	must(pmState.Status.UnmarshalText(pmStatus))

	return pmState
}

// SetPostMortemState sets the postmortem state of a process
func (s *ProcessStorage) SetPostMortemState(id uuid.UUID, pmState storage.PostMortemState) {
	tx := s.MustBegin()
	defer tx.MustCommit()

	pState := model.ProcessState{
		ProcessID: id,
		PMLevel:   pmState.Level,
		PMStatus:  pmState.Status,
		PMComment: pmState.Comment,
	}
	must(tx.Update(&pState,
		model.ProcessStatePMLevel,
		model.ProcessStatePMStatus,
		model.ProcessStatePMComment))
}

// EmitPosition implements storage.EmitPosition
type EmitPosition struct {
	tx             *TxHelper
	emission       model.Emission
	pos            storage.EnvelopeReadPos
	processStorage *ProcessStorage
	logentries     []storage.ProcessLogEntry
	lastError      error
}

// Get returns the position
func (p *EmitPosition) Get() (storage.EnvelopeReadPos, error) {
	return p.pos, p.lastError
}

// Set updates the position
func (p *EmitPosition) Set(pos storage.EnvelopeReadPos) error {
	if p.lastError != nil {
		return p.lastError
	}
	p.pos = pos
	p.emission.Start = pos.Start
	p.emission.Complete = pos.Complete
	b, err := json.Marshal(pos.Events)
	p.lastError = err
	must(err)
	p.emission.EventPositions = string(b)

	p.lastError = p.tx.Update(
		&p.emission,
		model.EmissionStart,
		model.EmissionComplete,
		model.EmissionEventPositions)
	must(p.lastError)
	logentry, err := p.processStorage.processLogEmission(
		p.tx,
		p.emission.ProcessID,
		storage.Emission{
			ProcessID:   p.emission.ProcessID,
			EnvelopeID:  p.emission.EnvelopeID,
			TargetActor: p.emission.TargetActor,
			TargetInputRef: storage.NodeInputRef{
				NodeID: p.emission.TargetNodeID,
				Input:  p.emission.TargetInput,
			},
		},
		pos)
	p.lastError = err
	must(err)
	p.logentries = append(p.logentries, logentry)
	return nil
}

// Unlock releases the lock on the position
func (p *EmitPosition) Unlock() error {
	if p.lastError != nil || len(p.logentries) == 0 {
		p.tx.Rollback()
		return nil
	}
	defer p.tx.RollbackIfOpened()
	return p.processStorage.updateStateAndCommit(
		p.tx, p.emission.ProcessID, p.logentries)
}
