package xbus

import (
	"bytes"
	"crypto/x509"
	"encoding/pem"
	"fmt"
	"io/ioutil"
	"os"
	"path"
	"strings"

	"github.com/m4rw3r/uuid"
	"github.com/nats-io/go-nats"
	"github.com/spf13/pflag"
	"github.com/spf13/viper"
	"gopkg.in/yaml.v2"

	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/go-xbus/tools"
)

const (
	confKeyAccountID   = "account-id"
	confKeyAccountName = "account-name"
	confKeyNatsURL     = "nats-url"
	confKeyForceStart  = "force-start"

	confKeyPersistentStore = "persistent-store"

	confKeyPrivateKey  = "tls.private-key"
	confKeyCertificate = "tls.certificate"
	confKeyCSR         = "tls.csr"
	confKeyRootCA      = "tls.rootca"
)

func fileExists(path string) bool {
	if _, err := os.Stat(path); err != nil {
		if os.IsNotExist(err) {
			return false
		}
		panic(err)
	}
	return true
}

func readPEM(blockType string, pemOrPath string) []byte {
	var (
		bytes []byte
		err   error
	)
	if strings.Contains(pemOrPath, "-----BEGIN") {
		if !strings.Contains(pemOrPath, "-----BEGIN "+blockType+"-----") {
			panic(fmt.Sprintf(
				"Invalid pem block type. Expected %s, got:\n%s",
				blockType, pemOrPath))
		}
		bytes = []byte(pemOrPath)
	} else if fileExists(pemOrPath) {
		bytes, err = ioutil.ReadFile(pemOrPath)
		if err != nil {
			panic(err)
		}
	} else {
		panic("Not a file nor a PEM block: " + pemOrPath)
	}
	block, _ := pem.Decode(bytes)
	if block == nil {
		panic("Could not decode PEM data")
	}
	return block.Bytes
}

// COptionsFromViper prepare a COptions from a Viper conf
// If writeback is true and the viper backend allows it, the
// option changes are wrote back to the configuration backend (
// most likely a configuration file).
func COptionsFromViper(
	v *viper.Viper,
	accountType api.AccountType,
	logger Logger,
	configDir string,
	configName string,
	writeback bool) COptions {
	opts := COptions{
		Name:       v.GetString(confKeyAccountName),
		Type:       accountType,
		ForceStart: v.GetBool(confKeyForceStart),
		ConnOptions: nats.Options{
			Url: v.GetString(confKeyNatsURL),
			AsyncErrorCB: func(c *nats.Conn, s *nats.Subscription, err error) {
				if s != nil {
					logger.Errorf("Error on sub %s: %s", s.Subject, err)
				} else {
					logger.Errorf("Error %s", err)
				}
			},
		},
	}

	if s := v.GetString(confKeyAccountID); s != "" {
		opts.ID = uuid.MustFromString(s)
	}

	for i := 0; ; i++ {
		prefix := fmt.Sprintf("actors.%d.", i)
		name := v.GetString(prefix + "name")
		if name == "" {
			break
		}
		var id uuid.UUID
		if s := v.GetString(prefix + "id"); s != "" {
			id = uuid.MustFromString(s)
		}
		var aKind api.ActorKind
		if s := v.GetString(prefix + "kind"); s != "" {
			if err := aKind.UnmarshalText([]byte(s)); err != nil {
				panic(err)
			}
		}
		var aStatus api.ActorStatus
		if s := v.GetString(prefix + "status"); s != "" {
			if err := aStatus.UnmarshalText([]byte(s)); err != nil {
				panic(err)
			}
		}
		roles := v.GetStringSlice(prefix + "roles")
		serviceType := v.GetString(prefix + "service.type")
		serviceSettings := v.GetStringMap(prefix + "service.settings")

		opts.ActorList = append(opts.ActorList, ActorOptions{
			Actor: api.Actor{
				ID:     id,
				Name:   name,
				Kind:   aKind,
				Status: aStatus,
				Roles:  roles,
			},
			Service: ServiceOptions{
				Type:     serviceType,
				Settings: serviceSettings,
			},
		})
	}

	if s := v.GetString(confKeyPrivateKey); s != "" {
		logger.Debugf("Loading pkey")
		key, err := x509.ParseECPrivateKey(readPEM("EC PRIVATE KEY", s))
		if err != nil {
			panic(err)
		}
		opts.PrivKey = key
	}

	if s := v.GetString(confKeyCSR); s != "" {
		logger.Debugf("Loading CSR")
		opts.CSR = readPEM("CERTIFICATE REQUEST", s)
	}

	if s := v.GetString(confKeyCertificate); s != "" {
		logger.Debugf("Loading certificate")
		opts.Certificate = readPEM("CERTIFICATE", s)
	}

	if s := v.GetString(confKeyRootCA); s != "" {
		logger.Debugf("Loading RootCA")

		opts.RootCAs = [][]byte{
			readPEM("CERTIFICATE", s),
		}
	}

	{
		store, err := NewFilePersistentStore(configDir + "/pstore.db")
		if err != nil {
			panic(err)
		}
		opts.PersistentStore = store
	}

	if writeback {
		cfgFile := v.ConfigFileUsed()
		if cfgFile == "" {
			cfgFile = path.Join(configDir, configName+".yaml")
		}
		opts.ChangeCallback = func(opts COptions) {
			confMap := v.AllSettings()
			confMap[confKeyAccountID] = opts.ID
			confMap[confKeyAccountName] = opts.Name
			confMap[confKeyNatsURL] = opts.ConnOptions.Url

			if opts.PrivKey != nil {
				if s := v.GetString(confKeyPrivateKey); s == "" {
					b, err := tools.PKeyToPEM(opts.PrivKey)
					if err != nil {
						panic(err)
					}
					confMap[confKeyPrivateKey] = string(b)
				} else {
					confMap[confKeyPrivateKey] = s
				}
			}
			if opts.CSR != nil {
				if s := v.GetString(confKeyCSR); s == "" {
					b := pem.EncodeToMemory(&pem.Block{
						Type:  "CERTIFICATE REQUEST",
						Bytes: opts.CSR,
					})
					confMap[confKeyCSR] = string(b)
				} else {
					confMap[confKeyCSR] = s
				}
			}
			if opts.Certificate != nil {
				if s := v.GetString(confKeyCertificate); s == "" {
					b := pem.EncodeToMemory(&pem.Block{
						Type:  "CERTIFICATE",
						Bytes: opts.Certificate,
					})
					confMap[confKeyCertificate] = string(b)
				} else {
					confMap[confKeyCertificate] = s
				}
			}
			if opts.RootCAs != nil {
				if s := v.GetString(confKeyRootCA); s == "" {
					var buf bytes.Buffer
					for _, cert := range opts.RootCAs {
						pem.Encode(&buf, &pem.Block{
							Type:  "CERTIFICATE",
							Bytes: cert,
						})
					}
					confMap[confKeyRootCA] = buf.String()
				} else {
					confMap[confKeyRootCA] = s
				}
			}
			if len(opts.ActorList) != 0 {
				actors := map[string]ActorOptions{}
				for i, actor := range opts.ActorList {
					actors[fmt.Sprintf("%d", i)] = actor
				}
				confMap["actors"] = actors
			}
			b, err := yaml.Marshal(confMap)
			if err != nil {
				panic(err)
			}
			logger.Noticef("Saving configuration to: %s", cfgFile)
			dirName := path.Dir(cfgFile)
			if _, err = os.Stat(dirName); err != nil {
				if os.IsNotExist(err) {
					logger.Debugf("Create %s", dirName)
					err = os.MkdirAll(dirName, 0700)
				}
				if err != nil {
					panic(err)
				}
			}
			if err := ioutil.WriteFile(cfgFile, b, 0600); err != nil {
				panic(err)
			}
		}
	}
	return opts
}

// DefineAndBindPFlags populates a viper.Viper and pflags.FlagSet
// with all the options suitable for a xbus client
func DefineAndBindPFlags(v *viper.Viper, flags *pflag.FlagSet) {
	flags.StringP(confKeyNatsURL, "u", "nats://localhost:4222", "NATS url")
	v.BindPFlag(confKeyNatsURL, flags.Lookup(confKeyNatsURL))

	flags.String(confKeyAccountID, "", "Account ID")
	v.BindPFlag(confKeyAccountID, flags.Lookup(confKeyAccountID))

	flags.String(confKeyAccountName, "", "Account Name")
	v.BindPFlag(confKeyAccountName, flags.Lookup(confKeyAccountName))

	flags.String(confKeyPrivateKey, "", "Private key (a path or a PEM encoded key)")
	v.BindPFlag(confKeyPrivateKey, flags.Lookup(confKeyPrivateKey))

	flags.String(confKeyCertificate, "", "Certificate (a path or a PEM encoded cert)")
	v.BindPFlag(confKeyCertificate, flags.Lookup(confKeyCertificate))

	flags.String(confKeyCSR, "", "CSR (a path or a PEM encoded CSR)")
	v.BindPFlag(confKeyCSR, flags.Lookup(confKeyCSR))

	flags.Bool(
		confKeyForceStart, false,
		"Enable force start (stay up even if the server is not reachable on start)")
	v.BindPFlag(confKeyForceStart, flags.Lookup(confKeyForceStart))
}
